==> kite-1.0a4.tar.gz: kite-1.0a4/modules/System/os/posix.c <==
/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include "kite_object.h"
#include "vm/kite_vm.h"

#define SET_CONSTANT(c) \
    kite_add_property(thread, newclass, # c, 1); \
    kite_set_property(thread, newclass, # c, kite_new_integer(thread, c));

/*****************************************************************************
 * Set signal handler
 ****************************************************************************/
KITE_CLASS_METHOD(Posix_setSignalHandler) {
    kite_object_t *sig, *method;

    KITE_GET_METHOD_ARGUMENT(sig, 1);
    KITE_GET_METHOD_ARGUMENT(method, 2);

    if (!KITE_IS_TYPE(sig, OBJ_INTEGER) || 
        (!KITE_IS_TYPE(method, OBJ_METHOD) && !KITE_IS_TYPE(method, OBJ_NULL))) { 
        kite_object_t *exc = 
            kite_new_exception(thd, "System.exceptions.TypeMismatch",                
                               "Arguments must be of type (integer, method/null)");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); 
    }

    if (KITE_GET_INTEGER(sig) >= NSIG || KITE_GET_INTEGER(sig) < 1) {
        kite_object_t *exc = 
            kite_new_exception(thd, "System.exceptions.InvalidArgument",                
                               "Signal number is invalid.");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); 
    }

    if (KITE_IS_TYPE(method, OBJ_NULL)) {
        signal(KITE_GET_INTEGER(sig), SIG_DFL);
        kite_object_t *oldsig = kite_signal_handlers[KITE_GET_INTEGER(sig)];
        kite_signal_handlers[KITE_GET_INTEGER(sig)] = NULL;
        oldsig->parent = NULL; 
        kite_dereference_object(oldsig);
    } else {
        signal(KITE_GET_INTEGER(sig), kite_handle_signal);
        kite_signal_handlers[KITE_GET_INTEGER(sig)] = method;
        method->parent = thd->vm->root_package;
    }

    kite_dereference_object(method);
    kite_dereference_object(sig);
}

/*****************************************************************************
 * Send ourselves a signal
 ****************************************************************************/
KITE_CLASS_METHOD(Posix_raise) {
    kite_object_t *sig;

    KITE_GET_METHOD_ARGUMENT(sig, 1);

    if (!KITE_IS_TYPE(sig, OBJ_INTEGER)) {
        kite_object_t *exc = 
            kite_new_exception(thd, "System.exceptions.TypeMismatch",                
                               "Argument must be integer.");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); 
    }

    if (KITE_GET_INTEGER(sig) >= NSIG || KITE_GET_INTEGER(sig) < 1) {
        kite_object_t *exc = 
            kite_new_exception(thd, "System.exceptions.InvalidArgument",                
                               "Signal number is invalid.");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); 
    }

    raise(KITE_GET_INTEGER(sig));

    kite_dereference_object(sig);
}

KITE_MODULE_INITIALIZER {
    /*
     * Create new class object. inherits is typically NULL for packages
     * and System.object.
     */
    kite_object_t *newclass = kite_new_class(thread, NULL, "posix");
    
    /* Add methods common to all classes. */
    kite_add_method(thread, newclass, "setSignalHandler", 
        kite_new_method(thread, FUNC_COMPILED, Posix_setSignalHandler, 2));
    kite_add_method(thread, newclass, "raise", 
        kite_new_method(thread, FUNC_COMPILED, Posix_raise, 1));

    /* Valid signals */
    SET_CONSTANT(SIGHUP);
    SET_CONSTANT(SIGINT);
    SET_CONSTANT(SIGQUIT);
    SET_CONSTANT(SIGILL);
    SET_CONSTANT(SIGTRAP);
    SET_CONSTANT(SIGABRT);
    SET_CONSTANT(SIGFPE);
    SET_CONSTANT(SIGKILL);
    SET_CONSTANT(SIGBUS);
    SET_CONSTANT(SIGSEGV);
    SET_CONSTANT(SIGSYS);
    SET_CONSTANT(SIGPIPE);
    SET_CONSTANT(SIGALRM);
    SET_CONSTANT(SIGTERM);
    SET_CONSTANT(SIGURG);
    SET_CONSTANT(SIGSTOP);
    SET_CONSTANT(SIGTSTP);
    SET_CONSTANT(SIGCONT);
    SET_CONSTANT(SIGCHLD);
    SET_CONSTANT(SIGTTIN);
    SET_CONSTANT(SIGTTOU);
    SET_CONSTANT(SIGIO);
    SET_CONSTANT(SIGXCPU);
    SET_CONSTANT(SIGXFSZ);
    SET_CONSTANT(SIGVTALRM);
    SET_CONSTANT(SIGPROF);
    SET_CONSTANT(SIGWINCH);
    SET_CONSTANT(SIGUSR1);
    SET_CONSTANT(SIGUSR2);
    SET_CONSTANT(NSIG);

    /* Add to parent to complete loading. */
    kite_add_property(thread, parent, "posix", 1);
    kite_set_property(thread, parent, "posix", newclass);
}

==> kite_1.0.2.orig.tar.gz: kite-1.0.2/modules/System/os/posix.c <==
/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdlib.h>
#include <string.h>
#ifndef WIN32
#include <unistd.h>
#endif /* WIN32 */
#include <signal.h>
#include <errno.h>
#include "kite_object.h"
#include "vm/kite_vm.h"

#define SET_CONSTANT(c) \
    kite_add_property(thread, newclass, # c, 1, # c); \
    kite_set_property(thread, newclass, # c, kite_new_integer(thread, c));

#ifndef WIN32
/*****************************************************************************
 * Set signal handler
 ****************************************************************************/
KITE_CLASS_METHOD(Posix_setSignalHandler) {
    kite_object_t *sig, *method, *oldsig;

    (void)this;
    KITE_GET_METHOD_ARGUMENT(sig, 1);
    KITE_GET_METHOD_ARGUMENT(method, 2);

    if (!KITE_IS_TYPE(sig, OBJ_INTEGER) || 
        (!KITE_IS_TYPE(method, OBJ_METHOD) && !KITE_IS_TYPE(method, OBJ_NULL))) { 
        kite_object_t *exc = 
            kite_new_exception(thd, "System.exceptions.TypeMismatch",                
                               "Arguments must be of type (integer, method/null)");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); 
        return;
    }

    if (KITE_GET_INTEGER(sig) >= NSIG || KITE_GET_INTEGER(sig) < 1) {
        kite_object_t *exc = 
            kite_new_exception(thd, "System.exceptions.InvalidArgument",                
                               "Signal number is invalid.");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        return;
    }

    if (KITE_IS_TYPE(method, OBJ_NULL)) {
        signal(KITE_GET_INTEGER(sig), SIG_DFL);
        oldsig = kite_signal_handlers[KITE_GET_INTEGER(sig)];
        kite_signal_handlers[KITE_GET_INTEGER(sig)] = NULL;
        oldsig->parent = NULL; 
        kite_dereference_object(oldsig);
    } else {
        signal(KITE_GET_INTEGER(sig), kite_handle_signal);
        kite_signal_handlers[KITE_GET_INTEGER(sig)] = method;
        method->parent = thd->vm->root_package;
    }

    kite_dereference_object(method);
    kite_dereference_object(sig);
}

/*****************************************************************************
 * Send ourselves a signal
 ****************************************************************************/
KITE_CLASS_METHOD(Posix_raise) {
    kite_object_t *sig;

    (void)this;
    KITE_GET_METHOD_ARGUMENT(sig, 1);

    if (!KITE_IS_TYPE(sig, OBJ_INTEGER)) {
        kite_object_t *exc = 
            kite_new_exception(thd, "System.exceptions.TypeMismatch",                
                               "Argument must be integer.");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); 
        return;
    }

    if (KITE_GET_INTEGER(sig) >= NSIG || KITE_GET_INTEGER(sig) < 1) {
        kite_object_t *exc = 
            kite_new_exception(thd, "System.exceptions.InvalidArgument",                
                               "Signal number is invalid.");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); 
        return;
    }

#ifdef __APPLE__
    if (thd->vm->thread_list[0] == thd) raise(KITE_GET_INTEGER(sig));
    else pthread_kill(thd->thread, KITE_GET_INTEGER(sig));
#else
    raise(KITE_GET_INTEGER(sig));
#endif // __APPLE__

    kite_dereference_object(sig);
}

/*****************************************************************************
 * popen()
 ****************************************************************************/
KITE_CLASS_METHOD(Posix_popen) {
    kite_object_t *arg1 = args->builtin_data.listvalue.car;
    kite_object_t *arg2 = 
        args->builtin_data.listvalue.cdr->builtin_data.listvalue.car;
    kite_object_t *file = kite_string_object(thd, arg1);
    kite_object_t *mode = kite_string_object(thd, arg2);
    kite_object_t *exc;

    this = kite_new_instance(thd, 
        kite_dereference_and_load(thd, "System.file"));
    this->builtin_data.filevalue = 
        popen(file->builtin_data.stringvalue.string, 
              mode->builtin_data.stringvalue.string);
    if (this->builtin_data.filevalue == NULL) {
        char buf[1024];
        strerror_r(errno, buf, 1024);
        exc = 
            kite_new_exception(thd, "System.exceptions.FileError", buf);
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        return;
    }

    kite_dereference_object(file);
    kite_dereference_object(mode);
    kite_vm_push(thd, this);
}

/*****************************************************************************
 * pclose()
 ****************************************************************************/
KITE_CLASS_METHOD(Posix_pclose) {
    kite_object_t *arg1 = args->builtin_data.listvalue.car;

    (void)this;
    (void)thd;
    
    pclose(arg1->builtin_data.filevalue);
}
#endif

KITE_MODULE_INITIALIZER(posix) {
    /*
     * Create new class object. inherits is typically NULL for packages
     * and System.object.
     */
    kite_object_t *newclass = kite_new_class(thread, NULL, "posix");
    kite_set_docstring(newclass, "Common POSIX functions and constants.");

#ifndef WIN32
    /* Add methods common to all classes. */
    kite_add_method(thread, newclass, "setSignalHandler", 
        kite_new_method_compiled_with_docs(thread, Posix_setSignalHandler, 
                                  "Sets the signal handler for a given signal.", 2,
                                  "sig", "Signal number to set for (see docs for valid signals).",
                                  "f", "Method to use (null for default handling)."));
    kite_add_method(thread, newclass, "raise", 
        kite_new_method_compiled_with_docs(thread, Posix_raise, 
                                  "Raises the given signal.", 1,
                                  "sig", "Signal to raise."));
    kite_add_method(thread, newclass, "popen", 
        kite_new_method_compiled_with_docs(thread, Posix_popen, 
                                  "Opens a pipe to another process.", 2,
                                  "command", "Command to run.",
                                  "mode", "Handle mode (r = read, w = write)"));
    kite_add_method(thread, newclass, "pclose", 
        kite_new_method_compiled_with_docs(thread, Posix_pclose, 
                                  "Closes a pipe opened with popen.", 1,
                                  "pipe", "Pipe returned from popen()."));
            
    /* Valid signals */
    SET_CONSTANT(SIGHUP);
    SET_CONSTANT(SIGINT);
    SET_CONSTANT(SIGQUIT);
    SET_CONSTANT(SIGILL);
    SET_CONSTANT(SIGTRAP);
    SET_CONSTANT(SIGABRT);
    SET_CONSTANT(SIGFPE);
    SET_CONSTANT(SIGKILL);
    SET_CONSTANT(SIGBUS);
    SET_CONSTANT(SIGSEGV);
    SET_CONSTANT(SIGSYS);
    SET_CONSTANT(SIGPIPE);
    SET_CONSTANT(SIGALRM);
    SET_CONSTANT(SIGTERM);
    SET_CONSTANT(SIGURG);
    SET_CONSTANT(SIGSTOP);
    SET_CONSTANT(SIGTSTP);
    SET_CONSTANT(SIGCONT);
    SET_CONSTANT(SIGCHLD);
    SET_CONSTANT(SIGTTIN);
    SET_CONSTANT(SIGTTOU);
    SET_CONSTANT(SIGIO);
    SET_CONSTANT(SIGXCPU);
    SET_CONSTANT(SIGXFSZ);
    SET_CONSTANT(SIGVTALRM);
    SET_CONSTANT(SIGPROF);
    SET_CONSTANT(SIGWINCH);
    SET_CONSTANT(SIGUSR1);
    SET_CONSTANT(SIGUSR2);
    SET_CONSTANT(NSIG);
#endif

    /* Add to parent to complete loading. */
    kite_add_property(thread, parent, "posix", 1, NULL);
    kite_set_property(thread, parent, "posix", newclass);
}
